Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Kot.LinkedBufferSourcefirst b returns the first element of the buffer b, which must be nonempty. It is equivalent to fst (pop b).
last b returns the last element of the buffer b, which must be nonempty. It is equivalent to snd (eject b).